# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from latexrestricted import PathSecurityError
from .command_clean import clean_initial_temp
from .messages import Messages
from .restricted import MintedTempRestrictedPath
from .version import __version_info__




def config(*, md5: str, timestamp: str, debug: bool, messages: Messages, data: dict[str, str] | None = None):
    config_file_name = f'_{md5}.config.minted'

    clean_initial_temp(md5=md5)

    config_lines = []
    minted_executable_version = f'{__version_info__.major}.{__version_info__.minor}.{__version_info__.micro}'
    config_lines.append(rf'\xdef\minted@executable@version{{\detokenize{{{minted_executable_version}}}}}%')
    config_lines.append(rf'\xdef\minted@executable@timestamp{{\detokenize{{{timestamp}}}}}%')
    if data is not None:
        tex_timestamp: str = data['timestamp']
        config_lines.append(rf'\xdef\minted@config@timestamp{{\detokenize{{{tex_timestamp}}}}}%')

    for openout_root in MintedTempRestrictedPath.tex_openout_roots():
        try:
            with (openout_root / config_file_name).open('w', encoding='utf8') as config_file:
                if data is None:
                    config_file.write('\n'.join(config_lines))
                    return
                tex_cachedir: str = data['cachedir']
                tex_cachepath: str
                if MintedTempRestrictedPath(tex_cachedir).is_absolute() or openout_root == MintedTempRestrictedPath.tex_cwd():
                    tex_cachepath = tex_cachedir
                else:
                    try:
                        tex_cachepath = (openout_root / tex_cachedir).relative_to(MintedTempRestrictedPath.tex_cwd()).as_posix()
                    except ValueError:
                        tex_cachepath = (openout_root / tex_cachedir).as_posix()
                if tex_cachepath and not tex_cachepath.endswith('/'):
                    tex_cachepath += '/'
                config_lines.append(rf'\xdef\minted@config@cachepath{{\detokenize{{{tex_cachepath}}}}}%')
                config_file.write('\n'.join(config_lines))
                return
        except (PermissionError, PathSecurityError):
            continue



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024-2025, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from os import environ as os_environ
# ruff: noqa: E402
os_environ['PYDEVD_DISABLE_FILE_VALIDATION'] = '1'
import argparse
import shutil
import sys
import textwrap
from typing import Callable



class ArgParser(argparse.ArgumentParser):
    def __init__(self, *, prog: str):
        super().__init__(
            prog=prog,
            allow_abbrev=False,
            formatter_class=argparse.RawTextHelpFormatter
        )
        self._prog = prog
        self._command_subparsers = None
        self._command_help_dict = None

    def add_command(self, name: str, *, help: str, func: Callable[..., None]):
        if self._command_subparsers is None:
            self._command_subparsers = self.add_subparsers(dest='subparser_name')
        if self._command_help_dict is None:
            self._command_help_dict = {}
        self._command_help_dict[name] = help
        parser = self._command_subparsers.add_parser(name, help=help)
        parser.set_defaults(func=func)
        parser.add_argument('--timestamp', help='LaTeX compile timestamp', required=True)
        parser.add_argument('--debug', help='Keep temp files for debugging', action='store_true')
        parser.add_argument('md5', help=r'MD5 hash based on \jobname')

    def print_help(self):
        term_columns = shutil.get_terminal_size()[0]
        help_lines = []
        if self._command_help_dict:
            help_lines.append(f'usage: {self._prog} [-h] [--version] COMMAND [--debug] --timestamp TIME MD5')
        else:
            help_lines.append(f'usage: {self._prog} [-h] [--debug] --timestamp TIME MD5')
        help_lines.extend([
            '',
            'Python executable for the LaTeX minted package.',
            r'Designed to be launched from within LaTeX via \ShellEscape.',
            '',
            'positional arguments:',
        ])
        if self._command_help_dict:
            command_choices = ', '.join(k for k in sorted(self._command_help_dict))
            if len(command_choices) - 14 - 2 > term_columns:
                command_choices = ','.join(k for k in sorted(self._command_help_dict))
            help_lines.append(f'  {"COMMAND":<12}{{{command_choices}}}')
            for k, v in sorted(self._command_help_dict.items()):
                help_lines.append(
                    textwrap.fill(f'{" "*16}* {k}: {v}', subsequent_indent=' '*16, width=term_columns)
                )
        help_lines.extend([
           r'  MD5         MD5 hash of LaTeX \jobname',
            '  TIME        LaTeX compilation timestamp (YYYYDDMMHHMMSS)',
            '',
            'options:',
            '  --debug     Keep temp files for debugging',
            '  -h, --help  Show this help message and exit',
            "  --version   Show program's version number and exit\n" if self._command_help_dict else '',
            'Repository: https://github.com/gpoore/minted',
            'CTAN: https://ctan.org/pkg/minted',
            'PyPI: https://pypi.org/project/latexminted',
        ])
        print('\n'.join(help_lines), file=sys.stdout)




def main():
    parser = ArgParser(
        prog='latexminted',
    )
    parser.set_defaults(func=lambda **x: parser.print_help())


    def get_version():
        from .version import __version__
        from latex2pydata import __version__ as latex2pydata_version
        from latexrestricted import __version__ as latexrestricted_version
        from pygments import __version__ as pygments_version
        library_versions = ', '.join([
            f'latex2pydata {latex2pydata_version}',
            f'latexrestricted {latexrestricted_version}',
            f'pygments {pygments_version}',
        ])
        return '\n'.join([
            f'latexminted {__version__}',
            'Python executable for the LaTeX minted package',
            f'Libraries: {library_versions}',
            'Repository: https://github.com/gpoore/minted',
            'CTAN: https://ctan.org/pkg/minted',
            'PyPI: https://pypi.org/project/latexminted',
        ])

    parser.add_argument('--version', action='version', version=get_version())

    # Lazy imports for functions that are designed to work only within LaTeX
    # shell escape.  These require SELFAUTOLOC and/or TEXSYSTEM environment
    # variables, which are set by LaTeX.  Without the lazy import, these can
    # raise errors and prevent `--help` etc. from functioning when `main()` is
    # not running within LaTeX shell escape.
    def batch(**kwargs):
        from .command_batch import batch
        batch(**kwargs)

    def clean(**kwargs):
        from .command_clean import clean
        clean(**kwargs)

    def clean_config(**kwargs):
        from .command_clean import clean_config_temp
        clean_config_temp(**kwargs)

    def clean_temp(**kwargs):
        from .command_clean import clean_temp_except_errlog
        clean_temp_except_errlog(**kwargs)

    def config(**kwargs):
        from .command_config import config
        config(**kwargs)

    def highlight(**kwargs):
        from .command_highlight import highlight
        highlight(**kwargs)

    def styledef(**kwargs):
        from .command_styledef import styledef
        styledef(**kwargs)

    parser.add_command('batch', help='Batch process highlight, styledef, and clean', func=batch)
    parser.add_command('clean', help='Clean up temp files and unused cache files', func=clean)
    parser.add_command('cleanconfig', help='Clean up config temp file', func=clean_config)
    parser.add_command('cleantemp', help='Clean up temp files', func=clean_temp)
    parser.add_command('config', help='Detect configuration and save it to file', func=config)
    parser.add_command('highlight', help='Highlight code and save it to file', func=highlight)
    parser.add_command(
        'styledef', help='Generate highlighting style definition and save it to file', func=styledef
    )

    cmdline_args = parser.parse_args()

    func_keys = set(['md5', 'timestamp', 'debug'])
    func_args = {k: v for k, v in vars(cmdline_args).items() if k in func_keys}
    if cmdline_args.subparser_name in ('cleanconfig', 'cleantemp'):
        # Some functions don't need all arguments
        func_args.pop('timestamp')
    md5: str | None = func_args.get('md5')
    timestamp: str | None = func_args.get('timestamp')
    if md5 is None or timestamp is None:
        cmdline_args.func(**func_args)
        sys.exit()

    from .command_clean import clean_messages, paths_skipped_in_initial_temp_cleaning
    from .debug import debug_mv_data
    from .load_data import load_data
    from .messages import Messages
    from .restricted import latexminted_config
    clean_messages(md5=md5)
    messages = Messages(md5=md5)
    func_args['messages'] = messages

    # All commands but `config` must exit immediately in the event of errors.
    # `config` must proceed as far as possible.   The `config()` function must
    # run, since its output is used on the LaTeX side in determining whether
    # the `latexminted` executable can be located.  Note that `config()` is
    # designed to handle the possibility of missing data.

    if latexminted_config.config_error and cmdline_args.subparser_name != 'config':
        messages.append_error(
            f'Failed to load latexminted configuration:  {latexminted_config.config_error}'
        )
        messages.communicate()
        sys.exit(1)

    try:
        maybe_data = load_data(md5=md5, messages=messages, timestamp=timestamp, command=cmdline_args.subparser_name)
    except Exception as e:
        messages.append_error(
            rf'Failed due to unexpected error (see \detokenize{{"{messages.errlog_file_name}"}} if it exists)'
        )
        messages.append_errlog(e)
        if cmdline_args.subparser_name == 'config':
            if latexminted_config.config_error:
                messages.append_error(
                    f'Failed to load latexminted configuration:  {latexminted_config.config_error}'
                )
            config(**func_args)
        messages.communicate()
        sys.exit(1)

    if messages.has_errors():
        if cmdline_args.subparser_name == 'config':
            if latexminted_config.config_error:
                messages.append_error(
                    f'Failed to load latexminted configuration:  {latexminted_config.config_error}'
                )
            config(**func_args)
        messages.communicate()
        sys.exit(1)
    if maybe_data is None:
        messages.append_error('Unexpectedly received no data without any error messages')
        if cmdline_args.subparser_name == 'config':
            if latexminted_config.config_error:
                messages.append_error(
                    f'Failed to load latexminted configuration:  {latexminted_config.config_error}'
                )
            config(**func_args)
        messages.communicate()
        sys.exit(1)

    data, data_path = maybe_data
    func_args['data'] = data
    debug: bool = func_args.get('debug', False)
    if debug:
        paths_skipped_in_initial_temp_cleaning.add(data_path)
    cmdline_args.func(**func_args)
    if debug:
        debug_mv_data(md5=md5, data_path=data_path)
    messages.communicate()
    if messages.has_errors():
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import hashlib
from pygments.lexer import Lexer
from latexrestricted import PathSecurityError
from ..err import CustomLexerError
from ._latexminted_config import latexminted_config
from ._restricted_pathlib import MintedTempRestrictedPath




def load_custom_lexer(lexer: str) -> type[Lexer]:
    if not latexminted_config.did_load_config_file:
        raise CustomLexerError('Missing ".latexminted_config"; custom lexers are disabled')

    lexer_path = MintedTempRestrictedPath(lexer)
    class_name: str = 'CustomLexer'
    if ':' in lexer_path.name:
        file_name, class_name = lexer_path.name.split(':', 1)
        lexer_path = lexer_path.parent / file_name
    if not lexer_path.name.lower().endswith('.py'):
        raise CustomLexerError(f'Custom lexer "{lexer}" is not a *.py file')

    try:
        lexer_bytes = lexer_path.read_bytes()
    except FileNotFoundError:
        raise CustomLexerError(f'Custom lexer "{lexer}" was not found')
    except PermissionError:
        raise CustomLexerError(f'Insufficient permission to open custom lexer "{lexer_path.as_posix()}"')
    except PathSecurityError as e:
        raise CustomLexerError(str(e))

    hasher = hashlib.sha256()
    hasher.update(lexer_bytes)
    hash = hasher.hexdigest()
    if not latexminted_config.is_custom_lexer_enabled(name=lexer_path.name, hash=hash):
        raise CustomLexerError(
            f'Custom lexer "{lexer_path.name}" is not enabled in ".latexminted_config"; '
            'check that SHA-256 hash is present and updated'
        )

    try:
        lexer_str = lexer_bytes.decode('utf-8-sig')
    except UnicodeDecodeError:
        raise CustomLexerError(f'Failed to decode custom lexer file "{lexer}" as UTF-8')

    namespace = {}
    try:
        exec(lexer_str, namespace)
    except Exception:
        raise CustomLexerError(f'Failed to exec custom lexer "{lexer}"; check the lexer file for errors')
    if class_name not in namespace:
        raise CustomLexerError(f'Custom lexer class "{class_name}" was not created by "{lexer}"')
    lexer_class = namespace[class_name]
    if not issubclass(lexer_class, Lexer):
        raise CustomLexerError(
            f'Custom lexer class "{class_name}" from "{lexer}" is not a subclass of the Pygments Lexer class'
        )
    return lexer_class



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import re
from ast import literal_eval
from collections import defaultdict
from json import loads as json_loads
from typing import Literal
try:
    from tomllib import loads as toml_loads
except ImportError:
    toml_loads = None
from latexrestricted import latex_config, PathSecurityError, ResolvedRestrictedPath
from ..err import LatexMintedConfigError




class LatexMintedConfigPath(ResolvedRestrictedPath):
    '''
    Class for `.latexminted_config` config files.  Only allows reading files
    that are not writable by LaTeX.
    '''
    def readable_dir(self):
        raise NotImplementedError

    def readable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        if self.name != '.latexminted_config':
            return (
                False,
                f'Cannot load config file "{self.as_posix()}"; only file name ".latexminted_config" is supported'
            )
        is_latex_writable, _ = ResolvedRestrictedPath(self).writable_file()
        if is_latex_writable:
            return (
                False,
                f'Cannot load config file "{self.as_posix()}" because LaTeX settings make it writable by LaTeX'
            )
        return (True, None)

    def writable_dir(self):
        raise NotImplementedError

    def writable_file(self):
        raise NotImplementedError


_file_extension_re = re.compile(r'\.[0-9A-Za-z]+')


class LatexMintedConfigSecurity(object):
    def __init__(self):
        self._enable_cwd_config: bool = False
        self._file_path_analysis: Literal['resolve'] | Literal['string'] = 'resolve'
        self._permitted_pathext_file_extensions: frozenset[str] | None = None

    @property
    def enable_cwd_config(self):
        return self._enable_cwd_config

    @property
    def file_path_analysis(self):
        return self._file_path_analysis

    @property
    def permitted_pathext_file_extensions(self):
        return self._permitted_pathext_file_extensions

    def update(self, **kwargs):
        enable_cwd_config = kwargs.pop('enable_cwd_config', None)
        if enable_cwd_config is not None:
            if enable_cwd_config not in (True, False):
                raise LatexMintedConfigError('"security.enable_cwd_config" must be boolean')
            self._enable_cwd_config = enable_cwd_config

        file_path_analysis = kwargs.pop('file_path_analysis', None)
        if file_path_analysis is not None:
            if file_path_analysis not in ('resolve', 'string'):
                raise LatexMintedConfigError('"security.file_path_analysis" must be "resolve" or "string"')
            self._file_path_analysis = file_path_analysis

        permitted_pathext = kwargs.pop('permitted_pathext_file_extensions', None)
        if permitted_pathext is not None:
            if not isinstance(permitted_pathext, list) or not all(isinstance(x, str) for x in permitted_pathext):
                raise LatexMintedConfigError(
                    '"security.permitted_pathext_file_extensions" must be a list of strings'
                )
            if not all(_file_extension_re.fullmatch(x) for x in permitted_pathext):
                raise LatexMintedConfigError(
                    '"security.permitted_pathext_file_extensions" must be a list of file extensions ".<ext>"'
                )
            self._permitted_pathext_file_extensions = frozenset(x.lower() for x in permitted_pathext)

        if kwargs:
            unknowns_keys = ', '.join(f'"{k}"' for k in kwargs)
            raise LatexMintedConfigError(f'"security" contains unknown keys {unknowns_keys}')


class LatexMintedConfig(object):
    def __init__(self, *, load_config_file: bool = True, config_error: LatexMintedConfigError | None = None):
        self._custom_lexers: dict[str, set[str]] = defaultdict(set)
        self._did_load_config_file: bool = False
        self._security: LatexMintedConfigSecurity = LatexMintedConfigSecurity()
        self._tex_cwd = LatexMintedConfigPath(latex_config.tex_cwd)

        self.config_error = config_error
        if load_config_file:
            if config_error:
                raise TypeError
            config_name = '.latexminted_config'
            self._load(LatexMintedConfigPath.home() / config_name)
            if latex_config.TEXMFHOME:
                self._load(LatexMintedConfigPath(latex_config.TEXMFHOME) / config_name)
            if self._security.enable_cwd_config:
                self._load(self._tex_cwd / config_name)

    def is_custom_lexer_enabled(self, *, name: str, hash: str):
        return hash.lower() in self._custom_lexers[name]

    @property
    def did_load_config_file(self):
        return self._did_load_config_file

    @property
    def security(self):
        return self._security


    _loaders = [json_loads, literal_eval]
    if toml_loads is not None:
        _loaders.append(toml_loads)

    def _load(self, path: LatexMintedConfigPath):
        try:
            data_str = path.read_text(encoding='utf-8-sig')
        except FileNotFoundError:
            return
        except PermissionError:
            raise LatexMintedConfigError(
                f'Permission error prevented access to config file "{path.as_posix()}"'
            )
        except PathSecurityError as e:
            if not path.is_file():
                return
            raise LatexMintedConfigError(str(e))
        except UnicodeDecodeError as e:
            raise LatexMintedConfigError(
                f'Failed to decode config file "{path.as_posix()}":  {e}'
            )
        if not data_str.strip():
            return

        for loader in self._loaders:
            try:
                data = loader(data_str)
                break
            except Exception:
                continue
        else:
            raise LatexMintedConfigError(
                f'Invalid config file "{path.as_posix()}":  could not deserialize data '
                '(check for syntax errors and similar issues)'
            )

        if not isinstance(data, dict) or not all(isinstance(k, str) for k in data):
            raise LatexMintedConfigError(
                f'Invalid config file "{path.as_posix()}":  data must be a dict with string keys'
            )

        custom_lexers = data.pop('custom_lexers', None)
        if custom_lexers:
            if not isinstance(custom_lexers, dict):
                raise LatexMintedConfigError(
                    f'Invalid config file "{path.as_posix()}":  "custom_lexers" must be a dict with string keys and '
                    'with values that are either strings or lists of strings'
                )
            for k, v in custom_lexers.items():
                if not isinstance(k, str):
                    raise LatexMintedConfigError(
                        f'Invalid config file "{path.as_posix()}":  "custom_lexers" must be a dict with string keys'
                    )
                if isinstance(v, str):
                    self._custom_lexers[k].add(v.lower())
                elif isinstance(v, list) and all(isinstance(x, str) for x in v):
                    self._custom_lexers[k].update(v_i.lower() for v_i in v)
                else:
                    raise LatexMintedConfigError(
                        f'Invalid config file "{path.as_posix()}":  "custom_lexers" must be a dict '
                        'with values that are either strings or lists of strings'
                    )

        security = data.pop('security', None)
        if security:
            if path.is_relative_to(self._tex_cwd):
                raise LatexMintedConfigError(
                    f'Invalid config file "{path.as_posix()}": "security" cannot be modified '
                    'by a config file in the TeX working directory'
                )
            if not (isinstance(security, dict) and all(isinstance(k, str) for k in security)):
                raise LatexMintedConfigError(
                    f'Invalid config file "{path.as_posix()}":  "security" must be a dict with string keys'
                )
            try:
                self._security.update(**security)
            except LatexMintedConfigError as e:
                raise LatexMintedConfigError(f'Invalid config file "{path.as_posix()}":  {e}')

        if data:
            unknowns_keys = ', '.join(f'"{k}"' for k in data)
            raise LatexMintedConfigError(
                f'Invalid config file "{path.as_posix()}":  unknown keys {unknowns_keys}'
            )

        self._did_load_config_file = True


try:
    latexminted_config = LatexMintedConfig()
except LatexMintedConfigError as e:
    latexminted_config = LatexMintedConfig(load_config_file=False, config_error=e)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from latex2pydata import loads as latex2pydata_loads
from typing import Any
from .messages import Messages
from .restricted import MintedTempRestrictedPath




def load_data(*, md5: str, messages: Messages, timestamp: str, command: str) -> tuple[list[dict[str, Any]] | dict[str, Any], MintedTempRestrictedPath] | None:
    data_file_name: str = f'_{md5}.data.minted'

    data_text: str | None = None
    for read_path in MintedTempRestrictedPath.tex_openout_roots()[:-1]:
        data_path = read_path / data_file_name
        try:
            data_text = data_path.read_text(encoding='utf8')
        except (FileNotFoundError, PermissionError):
            continue
        except UnicodeDecodeError:
            messages.append_error(rf'Failed to decode file \detokenize{{"{data_file_name}"}} (expected UTF-8)')
            return None
    if data_text is None:
        data_path = MintedTempRestrictedPath.tex_openout_roots()[-1] / data_file_name
        try:
            data_text = data_path.read_text(encoding='utf8')
        except FileNotFoundError:
            messages.append_error(rf'Failed to find file \detokenize{{"{data_file_name}"}}')
        except PermissionError:
            messages.append_error(rf'Insufficient permission to open file \detokenize{{"{data_file_name}"}}')
        except UnicodeDecodeError:
            messages.append_error(rf'Failed to decode file \detokenize{{"{data_file_name}"}} (expected UTF-8)')
    if data_text is None:
        return None

    try:
        data = latex2pydata_loads(data_text, schema={'cachefiles': 'list[str]'}, schema_missing='verbatim')
    except Exception as e:
        messages.append_error(
            rf'Failed to load data from file \detokenize{{"{data_file_name}"}} (see \detokenize{{"{messages.errlog_file_name}"}})'
        )
        messages.append_errlog(e)
        return None
    if isinstance(data, dict):
        if command != data['command']:
            messages.append_error(
                rf'''minted data file \detokenize{{"{data_file_name}"}} is for "{data['command']}", but expected "{command}"'''
            )
            return None
        if timestamp != data['timestamp']:
            messages.append_error(
                rf'minted data file \detokenize{{"{data_file_name}"}} has incorrect timestamp'
            )
            return None
    elif isinstance(data, list):
        if command != 'batch':
            messages.append_error(
                rf'''minted data file \detokenize{{"{data_file_name}"}} is for "batch", but expected "{command}"'''
            )
            return None
        valid_commands = set(['styledef', 'highlight', 'clean'])
        if not all(d['command'] in valid_commands for d in data):
            messages.append_error(
                rf'''minted data file \detokenize{{"{data_file_name}"}} is for "batch", but contains invalid data'''
            )
            return None
        if data and timestamp != data[0]['timestamp']:
            messages.append_error(
                rf'minted data file \detokenize{{"{data_file_name}"}} has incorrect timestamp'
            )
            return None
    else:
        messages.append_error(
            rf'minted data file \detokenize{{"{data_file_name}"}} contains unexpected data type "{type(data)}"'
        )
        return None
    return (data, data_path)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

from latexrestricted import PathSecurityError
from .restricted import MintedTempRestrictedPath




def debug_mv_data(*, md5, data_path: MintedTempRestrictedPath) -> MintedTempRestrictedPath | None:
    for n in range(1, 101):
        replacement_path = data_path.parent / f'_{md5}_{n}.data.minted'
        if not replacement_path.is_file():
            break
    else:
        return None
    try:
        data_path.replace(replacement_path)
    except (FileNotFoundError, PermissionError, PathSecurityError):
        return None
    return replacement_path



# -*- coding: utf-8 -*-
#
# Copyright (c) 2024, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the LaTeX Project Public License version 1.3c:
# https://www.latex-project.org/lppl.txt
#


from __future__ import annotations

import re
from typing import Literal
from latexrestricted import SafeWriteStringRestrictedPath, SafeWriteResolvedRestrictedPath
from ._latexminted_config import latexminted_config




# `[0-9a-zA-Z_-]+` covers temp file names `_<MD5 hash>` plus code cache file
# names `<MD5 hash>` plus style names `<style name>`.
_minted_temp_file_re = re.compile(r'[0-9a-zA-Z_-]+\.(?:config|data|errlog|highlight|index|message|style)\.minted')


if latexminted_config.security.file_path_analysis == 'resolve':
    MintedBaseRestrictedPath = SafeWriteResolvedRestrictedPath
elif latexminted_config.security.file_path_analysis == 'string':
    MintedBaseRestrictedPath = SafeWriteStringRestrictedPath
else:
    raise TypeError


class MintedTempRestrictedPath(MintedBaseRestrictedPath):
    '''
    Class for temp files, including cache files.
    '''
    def writable_file(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._writable_file_cache[self.cache_key]
        except KeyError:
            if _minted_temp_file_re.fullmatch(self.name):
                return super().writable_file()
            self._writable_file_cache[self.cache_key] = (
                False,
                f'file name "{self.name}" does not match the regular expression for minted temp and cache files'
            )
            return self._writable_file_cache[self.cache_key]


class MintedCodeRestrictedPath(MintedBaseRestrictedPath):
    '''
    Class for source code that is extracted from a document and saved in
    external files.
    '''
    if latexminted_config.security.permitted_pathext_file_extensions is not None:
        _prohibited_write_file_extensions = MintedBaseRestrictedPath.prohibited_write_file_extensions()
        if _prohibited_write_file_extensions is not None:
            _prohibited_write_file_extensions = frozenset(
                _prohibited_write_file_extensions - latexminted_config.security.permitted_pathext_file_extensions
            )

        @classmethod
        def prohibited_write_file_extensions(cls) -> frozenset[str] | None:
            return cls._prohibited_write_file_extensions

    def writable_dir(self) -> tuple[Literal[True], None] | tuple[Literal[False], str]:
        try:
            return self._writable_dir_cache[self.cache_key]
        except KeyError:
            # Require source code to be saved in a subdirectory of an output
            # root, rather than at the root itself.  If the code is not in a
            # location that is likely to be used as a current working
            # directory in a shell, then there is reduced potential for
            # unintentional code execution in the event that the user disables
            # other security settings.  This checks paths in both
            # raw and resolved forms to force increased security regardless of
            # `RestrictedPath` base class.
            if any(code_path == root_path for code_path in set([self, self.resolved()])
                   for root_path in self.tex_roots_with_resolved()):
                self._writable_dir_cache[self.cache_key] = (
                    False,
                    f'extracted code file "{self.as_posix()}" is not permitted in the TeX working directory, '
                    'TEXMF_OUTPUT_DIRECTORY, or TEXMFOUTPUT; it must be in a designated subdirectory'
                )
                return self._writable_dir_cache[self.cache_key]
            return super().writable_dir()
